#include "StdAfx.h"
#include "DateTimeEdit.h"

namespace FilterEdit
{
bool CDateTimeEdit::m_bInitialised = false;
CRegEx CDateTimeEdit::m_DateRegEx;
CRegEx CDateTimeEdit::m_TimeRegEx;
CRegEx CDateTimeEdit::m_DateTimeRegEx;
DWORD CDateTimeEdit::m_dwCenturyPrefix = 0;
DWORD CDateTimeEdit::m_dwTwoDigitYearMax = 0;
TCHAR CDateTimeEdit::m_chFieldOrder = 0;
int CDateTimeEdit::m_iDayIndexStart = -1;
int CDateTimeEdit::m_iDayIndexEnd = -1;
int CDateTimeEdit::m_iMonthIndexStart = -1;
int CDateTimeEdit::m_iMonthIndexEnd = -1;
int CDateTimeEdit::m_iYearIndexStart = -1;
int CDateTimeEdit::m_iYearIndexEnd = -1;
int CDateTimeEdit::m_iDateLength = -1;
TCHAR CDateTimeEdit::m_cDateSeparator = 0;
TCHAR CDateTimeEdit::m_cTimeSeparator = 0;
CString CDateTimeEdit::m_strDateToolTip;
CString CDateTimeEdit::m_strTimeToolTip;
CString CDateTimeEdit::m_strDateTimeToolTip;

// TODO:
// - Calculate calendar type automatically?  (Currently CAL_GREGORIAN)
CDateTimeEdit::CDateTimeEdit (const e_Mode eMode/* = eDate*/) :
	CBaseEdit (eMode == eDate ? &m_DateRegEx : eMode == eTime ? &m_TimeRegEx :
	&m_DateTimeRegEx),
	m_eMode (eMode),
	m_dwMargins (0),
	m_bFromSpin (false),
	m_pCalendarDlg (0),
	m_iHourIndexStart (-1),
	m_iHourIndexEnd (-1),
	m_iMinuteIndexStart (-1),
	m_iMinuteIndexEnd (-1),
	m_iSecondIndexStart (-1),
	m_iSecondIndexEnd (-1)
{
	if (!m_bInitialised)
	{
		TCHAR szData[4] = _T("");
		::GetLocaleInfo (LOCALE_USER_DEFAULT, LOCALE_SDATE, szData, 4);
		m_cDateSeparator = *szData;

#if _MSC_VER < 1310
		SYSTEMTIME st;

		::GetLocalTime (&st);
		m_dwCenturyPrefix = st.wYear + 24;
#else
		::GetCalendarInfo (LOCALE_USER_DEFAULT, CAL_GREGORIAN,
			CAL_RETURN_NUMBER | CAL_ITWODIGITYEARMAX, 0, 0,
			&m_dwCenturyPrefix);
#endif
		m_dwTwoDigitYearMax = m_dwCenturyPrefix % 1000;
		m_dwCenturyPrefix -= m_dwTwoDigitYearMax;
		m_dwCenturyPrefix /= 100;
		::GetLocaleInfo (LOCALE_USER_DEFAULT, LOCALE_IDATE, szData, 2);

		switch (*szData)
		{
			case '0':
				m_chFieldOrder = 'm';
				m_iMonthIndexStart = 0;
				m_iMonthIndexEnd = 2;
				m_iDayIndexStart = m_iMonthIndexEnd + 1;
				m_iDayIndexEnd = m_iDayIndexStart + 2;
				m_iYearIndexStart = m_iDayIndexEnd + 1;
				m_iYearIndexEnd = m_iYearIndexStart + 4;
				break;
			case '1':
				m_chFieldOrder = 'd';
				m_iDayIndexStart = 0;
				m_iDayIndexEnd = 2;
				m_iMonthIndexStart = m_iDayIndexEnd + 1;
				m_iMonthIndexEnd = m_iMonthIndexStart + 2;
				m_iYearIndexStart = m_iMonthIndexEnd + 1;
				m_iYearIndexEnd = m_iYearIndexStart + 4;
				break;
			case '2':
				m_chFieldOrder = 'y';
				m_iYearIndexStart = 0;
				m_iYearIndexEnd = 4;
				m_iMonthIndexStart = m_iYearIndexEnd + 1;
				m_iMonthIndexEnd = m_iMonthIndexStart + 2;
				m_iDayIndexStart = m_iMonthIndexEnd + 1;
				m_iDayIndexEnd = m_iDayIndexStart + 2;
				break;
			default:
				assert (0);
				break;
		}

		m_iDateLength = 10;
		::GetLocaleInfo (LOCALE_USER_DEFAULT, LOCALE_STIME, szData, 4);
		m_cTimeSeparator = *szData;

		const CString strDay (_T("(0[1-9]|[12]\\d|3[01])"));
		const CString strMonth (_T("(0[1-9]|1[0-2])"));
		const CString strYear (_T("\\d{4}"));
		const CString strYearTooltip (_T("YYYY"));
		const CString strMonthTooltip (_T("MM"));
		const CString strDayTooltip (_T("DD"));
		CString strEscTimeSeparator (m_cTimeSeparator);
		CString strEscDateSeparator (m_cDateSeparator);
		CString strTime;
		CString strRegEx;
		CString strFormat;

		EscapeString (strEscTimeSeparator);
		strTime = _T("([01][0-9]|2[0-3])(") + strEscTimeSeparator +
			_T("[0-5][0-9]){2}");
		EscapeString (strEscDateSeparator);

		if (m_chFieldOrder == 'y')
		{
			strRegEx = strYear + strEscDateSeparator + strMonth +
				strEscDateSeparator + strDay;
			strFormat = strYearTooltip;
			strFormat += m_cDateSeparator;
			strFormat += strMonthTooltip;
			strFormat += m_cDateSeparator;
			strFormat += strDayTooltip;
		}
		else
		{
			if (m_chFieldOrder == 'd')
			{
				strRegEx = strDay + strEscDateSeparator + strMonth +
					strEscDateSeparator + strYear;
				strFormat += strDayTooltip;
				strFormat += m_cDateSeparator;
				strFormat += strMonthTooltip;
			}
			else
			{
				strRegEx = strMonth + strEscDateSeparator + strDay +
					strEscDateSeparator + strYear;
				strFormat += strMonthTooltip;
				strFormat += m_cDateSeparator;
				strFormat += strDayTooltip;
			}

			strFormat += m_cDateSeparator;
			strFormat += strYearTooltip;
		}

		m_DateRegEx.assign (strRegEx);
		m_TimeRegEx.assign (strTime);
		m_DateTimeRegEx.assign (strRegEx + ' ' + strTime);
		m_strDateToolTip.Format (_T("Please enter a date in format %s"),
			strFormat);
		m_strTimeToolTip.Format (_T("Please enter a time in format HH:MM:SS"),
			strFormat);
		m_strDateTimeToolTip.Format (_T("Please enter a date and time in format ")
			_T("%s HH:MM:SS"), strFormat);
	}

	SetTimeIndexes ();
}

CDateTimeEdit::~CDateTimeEdit ()
{
}

void CDateTimeEdit::SetMode (const e_Mode eMode)
{
	ClearText ();
	m_eMode = eMode;

	switch(m_eMode)
	{
	case eDate:
		SetRegEx (&m_DateRegEx);
		SetToolTipText (m_strDateToolTip);
		break;
	case eTime:
		SetRegEx (&m_TimeRegEx);
		SetToolTipText (m_strTimeToolTip);
		break;
	case eDateTime:
		SetRegEx (&m_DateTimeRegEx);
		SetToolTipText (m_strDateTimeToolTip);
		break;
	default:
		assert (0);
		break;
	}

	SetTimeIndexes ();

	if (m_eMode == eTime)
	{
		m_ButtonCalendar.ShowWindow (FALSE);
		SetMargins (LOWORD (m_dwMargins), HIWORD (m_dwMargins));
	}
	else
	{
		CRect ClientRect;

		GetClientRect (ClientRect);
		SetMargins (LOWORD(m_dwMargins), HIWORD (m_dwMargins) +
			ClientRect.bottom - ClientRect.top);
		m_ButtonCalendar.ShowWindow (TRUE);
	}
}

CDateTimeEdit::e_Mode CDateTimeEdit::GetMode () const
{
	return m_eMode;
}

bool CDateTimeEdit::SetDateTime (const COleDateTime &dt)
{
	bool bSuccess = true;

	if (dt.GetStatus () == COleDateTime::valid)
	{
		switch (m_eMode)
		{
		case eDate:
			bSuccess = SetDate (dt.GetDay (), dt.GetMonth (), dt.GetYear ());
			break;
		case eTime:
			bSuccess = SetTime (dt.GetHour (), dt.GetMinute (),
				dt.GetSecond ());
			break;
		case eDateTime:
		{
			CString strText;

			switch (m_chFieldOrder)
			{
				case 'd':
					strText.Format (_T("%02d%c%02d%c%04d %02d%c%02d%c%02d"),
						dt.GetDay (),
						m_cDateSeparator, dt.GetMonth (), m_cDateSeparator,
						dt.GetYear (), dt.GetHour (), m_cTimeSeparator,
						dt.GetMinute (), m_cTimeSeparator, dt.GetSecond ());
					bSuccess = IsDayMonthYearValid (strText);
					break;
				case 'm':
					strText.Format (_T("%02d%c%02d%c%04d %02d%c%02d%c%02d"),
						dt.GetMonth (), m_cDateSeparator, dt.GetDay (),
						m_cDateSeparator, dt.GetYear (), dt.GetHour (),
						m_cTimeSeparator, dt.GetMinute (),
						m_cTimeSeparator, dt.GetSecond ());
					bSuccess = IsMonthDayYearValid (strText);
					break;
				case 'y':
					strText.Format (_T("%04d%c%02d%c%02d %02d%c%02d%c%02d"),
						dt.GetYear (), m_cDateSeparator, dt.GetMonth (),
						m_cDateSeparator, dt.GetDay (), dt.GetHour (),
						m_cTimeSeparator, dt.GetMinute (),
						m_cTimeSeparator, dt.GetSecond ());
					bSuccess = IsYearMonthDayValid (strText);
					break;
				default:
					assert (0);
					break;
			}

			if (bSuccess)
			{
				SetWindowText (strText);
			}

			break;
		}
		default:
			assert (0);
			break;
		}
	}
	else
	{
		ClearText ();
	}

	return bSuccess;
}

bool CDateTimeEdit::GetDateTime (COleDateTime &dt)
{
	bool bSuccess = IsInputValid (false);

	if (bSuccess)
	{
		switch (m_eMode)
		{
		case eDate:
			dt.SetDate (GetYear (), GetMonth (), GetDay ());
			break;
		case eTime:
			dt.SetTime (GetHour (), GetMinute (), GetSecond ());
			break;
		case eDateTime:
			dt.SetDateTime (GetYear (), GetMonth (), GetDay (), GetHour (),
				GetMinute (), GetSecond ());
			break;
		default:
			assert (0);
			break;
		}
	}
	else
	{
		dt.SetStatus (COleDateTime::null);
	}

	return bSuccess;
}

bool CDateTimeEdit::SetDate (const int iDay, const int iMonth, const int iYear)
{
	bool bSuccess = m_eMode != eTime;

	if (bSuccess)
	{
		CString strDate;

		switch (m_chFieldOrder)
		{
			case 'd':
				strDate.Format (_T("%02d%c%02d%c%04d"), iDay,
					m_cDateSeparator, iMonth, m_cDateSeparator, iYear);
				bSuccess = IsDayMonthYearValid (strDate);
				break;
			case 'm':
				strDate.Format (_T("%02d%c%02d%c%04d"), iMonth,
					m_cDateSeparator, iDay, m_cDateSeparator, iYear);
				bSuccess = IsMonthDayYearValid (strDate);
				break;
			case 'y':
				strDate.Format (_T("%04d%c%02d%c%02d"), iYear,
					m_cDateSeparator, iMonth, m_cDateSeparator, iDay);
				bSuccess = IsYearMonthDayValid (strDate);
				break;
			default:
				assert (0);
				break;
		}

		if (bSuccess)
		{
			if (m_eMode == eDateTime)
			{
				SetWindowText (strDate + _T(" 00:00:00"));
			}
			else
			{
				SetWindowText (strDate);
			}
		}
	}

	return bSuccess;
}

bool CDateTimeEdit::GetDate (int &iDay, int &iMonth, int &iYear)
{
	const bool bSuccess = m_eMode != eTime && IsInputValid (false);

	if (bSuccess)
	{
		iDay = GetDay ();
		iMonth = GetMonth ();
		iYear = GetYear ();
	}

	return bSuccess;
}

bool CDateTimeEdit::GetTimeStamp (SQL_TIMESTAMP_STRUCT *pTS)
{
	const bool bSuccess = IsInputValid (false);
	int iDay = -1;
	int iMonth = -1;
	int iYear = -1;
	int iHour = -1;
	int iMinute = -1;
	int iSecond = -1;

	switch (m_eMode)
	{
	case eDate:
		GetDate (iDay, iMonth, iYear);
		break;
	case eTime:
		GetTime (iHour, iMinute, iSecond);
		break;
	case eDateTime:
		GetDate (iDay, iMonth, iYear);
		GetTime (iHour, iMinute, iSecond);
		break;
	default:
		assert (0);
		break;
	}

	pTS->day = static_cast<SQLUSMALLINT>(iDay);
	pTS->month = static_cast<SQLUSMALLINT>(iMonth);
	pTS->year = static_cast<SQLSMALLINT>(iYear);
	pTS->hour = static_cast<SQLUSMALLINT>(iHour);
	pTS->minute = static_cast<SQLUSMALLINT>(iMinute);
	pTS->second = static_cast<SQLUSMALLINT>(iSecond);
	return bSuccess;
}

int CDateTimeEdit::GetDay ()
{
	int iDay = -1;

	if (m_eMode != eTime && IsInputValid (false))
	{
		CString strDate;

		GetWindowText (strDate);
		iDay = GetDay (strDate);
	}

	return iDay;
}

int CDateTimeEdit::GetMonth ()
{
	int iMonth = -1;

	if (m_eMode != eTime && IsInputValid (false))
	{
		CString strDate;

		GetWindowText (strDate);
		iMonth = GetMonth (strDate);
	}

	return iMonth;
}

int CDateTimeEdit::GetYear ()
{
	int iYear = -1;

	if (m_eMode != eTime && IsInputValid (false))
	{
		CString strDate;

		GetWindowText (strDate);
		iYear = GetYear (strDate);
	}

	return iYear;
}

bool CDateTimeEdit::SetTime (const int iHour, const int iMinute, const int iSecond)
{
	bool bSuccess = m_eMode != eDate && iHour >= 0 && iHour <= 23 &&
		iMinute >= 1 && iMinute <= 59 && iSecond >= 1 && iSecond <= 59;

	if (m_eMode == eDateTime)
	{
		// Date must be followed by a space:
		bSuccess = GetWindowTextLength () > m_iDateLength;
	}

	if (bSuccess)
	{
		CString strTime;

		strTime.Format (_T("%02d%c%02d%c%04d"), iHour,
			m_cTimeSeparator, iMinute, m_cTimeSeparator, iSecond);
		SetSel (m_iHourIndexStart, m_iSecondIndexEnd);
		ReplaceSel (strTime);
	}

	return bSuccess;
}

bool CDateTimeEdit::GetTime (int &iHour, int &iMinute, int &iSecond)
{
	bool bSuccess = m_eMode != eDate && IsInputValid (false);

	if (bSuccess)
	{
		iHour = GetHour ();
		iMinute = GetMinute ();
		iSecond = GetSecond ();
	}

	return bSuccess;
}

int CDateTimeEdit::GetHour ()
{
	int iHour = -1;

	if (m_eMode != eDate && IsInputValid (false))
	{
		CString strText;

		GetWindowText (strText);
		strText = strText.Mid (m_iHourIndexStart, 2);
		iHour = _tstoi (strText);
	}

	return iHour;
}

int CDateTimeEdit::GetMinute ()
{
	int iMinute = -1;

	if (m_eMode != eDate && IsInputValid (false))
	{
		CString strText;

		GetWindowText (strText);
		strText = strText.Mid (m_iMinuteIndexStart, 2);
		iMinute = _tstoi (strText);
	}

	return iMinute;
}

int CDateTimeEdit::GetSecond ()
{
	int iSecond = -1;

	if (m_eMode != eDate && IsInputValid (false))
	{
		CString strText;

		GetWindowText (strText);
		strText = strText.Mid (m_iSecondIndexStart, 2);
		iSecond = _tstoi (strText);
	}

	return iSecond;
}

void CDateTimeEdit::CreateToolTip (CWnd *pParent)
{
	switch (m_eMode)
	{
	case eDate:
		CBaseEdit::CreateToolTip (pParent, m_strDateToolTip);
		break;
	case eTime:
		CBaseEdit::CreateToolTip (pParent, m_strTimeToolTip);
		break;
	case eDateTime:
		CBaseEdit::CreateToolTip (pParent, m_strDateTimeToolTip);
		break;
	default:
		assert (0);
		break;
	}
}

// TODO: Review function
void CDateTimeEdit::Increment (const int iInc)
{
	const int iLen = GetWindowTextLength ();
	int iSelStart = 0;
	int iSelEnd = 0;

	// Note GetSel over-ridden
	GetSel (iSelStart, iSelEnd);

	int iStart = 0;
	int iEnd = 0;
	e_Part ePart = LocatePart (iSelStart, iStart, iEnd);

	if (iLen < iEnd)
	{
		MoveRight ();
	}
	else
	{
		CString strText;
		int i = 0;
		CString strPart;
		bool bProcess = true;

		GetWindowText (strText);
		i = _tstoi (strText.Mid (iStart, iEnd - iStart));
		i += iInc > 0 ? 1 : -1;

		if (ePart > eYear)
		{
			if (i < 0)
			{
				switch (ePart)
				{
				case eHour:
					i = 23;
					break;
				case eMinute:
				case eSecond:
					i = 59;
					break;
				default:
					assert (0);
					break;
				}
			}
			else
			{
				switch (ePart)
				{
				case eHour:
					if (i > 23) i = 0;
					break;
				case eMinute:
				case eSecond:
					if (i > 59) i = 0;
					break;
				default:
					assert (0);
					break;
				}
			}

			strPart.Format (_T("%02d"), i);
		}
		else
		{
			if (i < 1)
			{
				switch (ePart)
				{
				case eDay:
					i = MaxDay (GetMonth (strText), GetYear (strText));
					break;
				case eMonth:
					// December has 31 days, so no need to check day.
					i = 12;
					break;
				case eYear:
					// Handled in else below.
					break;
				default:
					assert (0);
					break;
				}
			}
			else
			{
				switch (ePart)
				{
				case eDay:
					// Handled below
					break;
				case eMonth:
				{
					const int iDay = GetDay (strText);

					// Select a valid month, wrapping around if necessary.
					while (!ValidateDayMonth (iDay, i))
					{
						i += iInc;

						if (i == 0) i = 12;
						else if (i == 13) i = 1;
					}

					// Check for leap year.
					if (!ValidateLeapYear (iDay, i, GetYear (strText)))
					{
						i += iInc;
					}

					break;
				}
				case eYear:
					// Check for leap year.
					if (GetDay (strText) == 29 && GetMonth (strText) == 2)
					{
						while (!ValidateLeapYear (29, 2, i))
						{
							i += iInc;

							// Check for under/overflow.
							if (i < 1753) i = 9996;
							else if (i > 9999) i = 1756;
						}
					}
					else
					{
						if (i < 1753)
						{
							i = 9999;
						}
						else if (i > 9999)
						{
							i = 1753;
						}
					}

					break;
				}
			}

			if (ePart == eYear)
			{
				strPart.Format (_T("%04d"), i);
			}
			else
			{
				strPart.Format (_T("%02d"), i);
			}

			strText.Delete (iStart, iEnd - iStart);
			strText.Insert (iStart, strPart);
			bProcess = IsDateValid (strText);

			if (!bProcess && ePart != eYear)
			{
				strPart = "01";
			}
		}

		// Needed when up arrow pressed with nothing selected.
		SetSel (iStart, iEnd);
		ReplaceSel (strPart, TRUE);
	}

	SetSel (iStart, iEnd);
	m_bFromSpin = true;
}

void CDateTimeEdit::GetSel (int &nStartChar, int &nEndChar) const
{
	CEdit::GetSel (nStartChar, nEndChar);

	if (nEndChar < nStartChar)
	{
		std::swap (nStartChar, nEndChar);
	}
}

bool CDateTimeEdit::SyntaxCheck (const CString &strText)
{
	bool bSuccess = true;
	CString strTemp (strText);
	bool bChanged = m_eMode != eTime ? ExpandYear (strTemp) : false;

	if (m_eMode != eDate)
	{
		bChanged |= FormatTime (strTemp);
	}

	if (bChanged)
	{
		const bool bProcessChars = GetProcessChars ();

		SetProcessChars (false);
		SetWindowText (strText);
		SetProcessChars (bProcessChars);
	}

	bSuccess = CBaseEdit::SyntaxCheck (strTemp);
	return bSuccess;
}

bool CDateTimeEdit::SemanticCheck (const CString &strText,
	CString *pstrErrorMsg/* = 0*/)
{
	return m_eMode == eTime ? true :
		IsDateValid (strText, pstrErrorMsg);
}

void CDateTimeEdit::FormatDateTime ()
{
	CString strText;
	bool bFormatted = false;

	GetWindowText (strText);

	if (m_eMode != eTime)
	{
		bFormatted = FormatDate (strText);
	}

	if (m_eMode != eDate)
	{
		bFormatted |= FormatTime (strText);
	}

	if (bFormatted)
	{
		const bool bProcessChars = GetProcessChars ();

		SetProcessChars (false);
		SetWindowText (strText);
		SetProcessChars (bProcessChars);
	}
}

bool CDateTimeEdit::FormatDate (CString &strDate)
{
	bool bFormatted = false;

	bFormatted = FormatDay (strDate);
	bFormatted |= FormatMonth (strDate);
	bFormatted |= FormatYear (strDate);

	return bFormatted;
}

bool CDateTimeEdit::FormatDay (CString &strText)
{
	bool bFormatted = false;

	if (m_eMode != eTime)
	{
		if (strText.GetLength () >= m_iDayIndexEnd)
		{
			const TCHAR ch = strText[m_iDayIndexEnd - 1];

			if (ch < '0' || ch > '9')
			{
				if (strText[m_iDayIndexStart] == '0')
				{
					strText.SetAt (m_iDayIndexStart, '1');
				}

				strText.Insert (m_iDayIndexStart, '0');
				bFormatted = true;
			}
		}

		if (m_chFieldOrder != 'y')
		{
			if (strText.GetLength () == m_iDayIndexEnd)
			{
				strText += m_cDateSeparator;
				bFormatted = true;
			}
			else if (strText.GetLength () >= m_iDayIndexEnd &&
				strText[m_iDayIndexEnd] != m_cDateSeparator)
			{
				strText.Insert (m_iDayIndexEnd, m_cDateSeparator);
				bFormatted = true;
			}
		}
	}

	return bFormatted;
}

bool CDateTimeEdit::FormatMonth (CString &strText)
{
	bool bFormatted = false;

	if (strText.GetLength () >= m_iMonthIndexEnd)
	{
		const TCHAR ch = strText[m_iMonthIndexEnd - 1];

		if (ch < '0' || ch > '9')
		{
			if (strText[m_iMonthIndexStart] == '0')
			{
				strText.SetAt (m_iMonthIndexStart, '1');
			}

			strText.Insert (m_iMonthIndexStart, '0');
			bFormatted = true;
		}

		if (m_chFieldOrder != 'y')
		{
			if (strText.GetLength () == m_iMonthIndexEnd)
			{
				strText += m_cDateSeparator;
				bFormatted = true;
			}
			else if (strText[m_iMonthIndexEnd] != m_cDateSeparator)
			{
				strText.Insert (m_iMonthIndexEnd, m_cDateSeparator);
				bFormatted = true;
			}
		}
	}

	return bFormatted;
}

bool CDateTimeEdit::FormatYear (CString &strText)
{
	bool bFormatted = false;

	if (strText.GetLength () >= m_iYearIndexEnd)
	{
		TCHAR ch = strText[m_iYearIndexStart + 1];

		if (ch < '0' || ch > '9')
		{
			strText.Insert (m_iYearIndexStart, '0');
			bFormatted = true;
		}

		// Allow previous if to run to create two digits.
		ch = strText[m_iYearIndexStart + 2];

		if (ch < '0' || ch > '9')
		{
			const DWORD dwYear = _tstoi (strText.Left (2));
			DWORD dwCenturyPrefix = m_dwCenturyPrefix;
			CString strPrefix;

			if (dwYear > m_dwTwoDigitYearMax)
			{
				dwCenturyPrefix--;
			}

			strPrefix.Format (_T("%d"), dwCenturyPrefix);
			strText.Insert (m_iYearIndexStart, strPrefix);
			bFormatted = true;
		}

		ch = strText[m_iYearIndexStart + 3];

		if (ch < '0' || ch > '9')
		{
			strText.Insert (m_iYearIndexStart, static_cast<TCHAR>
				(m_dwCenturyPrefix / 10) + '0');

			if (_tstoi (strText.Mid (m_iYearIndexStart, 4)) < 1752)
			{
				strText.Delete (m_iYearIndexStart, 4);
				strText.Insert (m_iYearIndexStart, _T("1753"));
			}

			bFormatted = true;
		}
	}

	if (m_eMode == eDateTime && strText.GetLength () == m_iDateLength)
	{
		strText += ' ';
		bFormatted = true;
	}

	return bFormatted;
}

bool CDateTimeEdit::FormatTime (CString &strText)
{
	bool bFormatted = false;
	int iFirstIndex = strText.Find (m_cTimeSeparator);
	int iSecondIndex = -1;

	if (iFirstIndex > 0)
	{
		if (iFirstIndex < m_iHourIndexEnd)
		{
			strText.Insert (m_iHourIndexStart, '0');
			iFirstIndex++;
			bFormatted = true;
		}

		iSecondIndex = strText.Find (m_cTimeSeparator, iFirstIndex + 1);

		if (iSecondIndex > 0)
		{
			if (iSecondIndex < m_iMinuteIndexEnd)
			{
				strText.Insert (m_iMinuteIndexStart, '0');
				bFormatted = true;
			}
		}
	}

	if (m_eMode == eDateTime && strText.GetLength () == m_iDateLength)
	{
		strText += ' ';
		bFormatted = true;
	}
	else if (strText.GetLength () == m_iHourIndexEnd ||
		strText.GetLength () == m_iMinuteIndexEnd)
	{
		strText += m_cTimeSeparator;
		bFormatted = true;
	}

	return bFormatted;
}

bool CDateTimeEdit::IsDateValid (const CString &strText,
	CString *pstrErrorMsg/* = 0*/)
{
	bool bValid = true;

	switch (m_chFieldOrder)
	{
		case 'd':
			bValid = IsDayMonthYearValid (strText, pstrErrorMsg);
			break;
		case 'm':
			bValid = IsMonthDayYearValid (strText, pstrErrorMsg);
			break;
		case 'y':
			bValid = IsYearMonthDayValid (strText, pstrErrorMsg);
			break;
		default:
			assert (0);
			bValid = false;
			break;
	}

	return bValid;
}

bool CDateTimeEdit::IsDayMonthYearValid (const CString &strDate,
	CString *pstrErrorMsg/* = 0*/)
{
	bool bValid = true;
	const int iLen = strDate.GetLength ();
	int iDay = 0;
	int iMonth = 0;
	int iYear = 0;

	if (iLen >= 1)
	{
		if (iLen >= 2)
		{
			iDay = _tstoi (strDate.Left (2));
		}
		else
		{
			iDay = _tstoi (strDate);
		}

		// Day only test
		bValid = iLen == 1 || (iDay >= 1 && iDay <= 31);

		if (!bValid)
		{
			if (pstrErrorMsg)
				*pstrErrorMsg = _T("Day must be 1-31");
		}
		else if (iLen > m_iMonthIndexStart)
		{
			if (iLen >= m_iMonthIndexEnd)
			{
				iMonth = _tstoi (strDate.Mid (m_iMonthIndexStart, 2));
			}
			else
			{
				iMonth = _tstoi (strDate.Mid (m_iMonthIndexStart));
			}

			// Month only test
			bValid = iLen == m_iMonthIndexEnd - 1 || (iMonth >= 1 && iMonth <= 12);

			if (bValid)
			{
				// Day and Month test
				bValid = ValidateDayMonth (iDay, iMonth);

				if (bValid)
				{
					if (iLen >= m_iYearIndexEnd)
					{
						iYear = _tstoi (strDate.Mid (m_iYearIndexStart, 4));
						bValid = ValidateLeapYear (iDay, iMonth, iYear);

						if (bValid)
						{
							bValid = iYear > 1752;

							if (!bValid && pstrErrorMsg)
								*pstrErrorMsg = _T("Years < 1753 not supported.");
						}
						else if (pstrErrorMsg)
							*pstrErrorMsg = _T("Invalid Leap Year");
					}
				}
				else
				{
					if (pstrErrorMsg)
						*pstrErrorMsg = _T("Invalid Day/Month Combination");
				}
			}
			else
			{
				if (pstrErrorMsg)
					*pstrErrorMsg = _T("Month must be 1-12");
			}
		}
	}

	return bValid;
}

bool CDateTimeEdit::IsMonthDayYearValid (const CString &strDate,
	CString *pstrErrorMsg/* = 0*/)
{
	bool bValid = true;
	const int iLen = strDate.GetLength ();
	int iDay = 0;
	int iMonth = 0;
	int iYear = 0;

	if (iLen >= 1)
	{
		if (iLen >= 2)
		{
			iMonth = _tstoi (strDate.Left (2));
		}
		else
		{
			iMonth = _tstoi (strDate);
		}

		// Month only test
		bValid = iLen == 1 || (iMonth >= 1 && iMonth <= 12);

		if (!bValid)
		{
			if (pstrErrorMsg)
				*pstrErrorMsg = _T("Month must be 1-12");
		}
		else if (iLen > m_iDayIndexStart)
		{
			if (iLen >= m_iDayIndexEnd)
			{
				iDay = _tstoi (strDate.Mid (m_iDayIndexStart, 2));
			}
			else
			{
				iDay = _tstoi (strDate.Mid (m_iDayIndexStart));
			}

			// Day only test
			bValid = iLen == m_iDayIndexEnd - 1 || (iDay >= 1 && iDay <= 31);

			if (bValid)
			{
				// Day and Month test
				bValid = ValidateDayMonth (iDay, iMonth);

				if (bValid)
				{
					if (iLen >= m_iYearIndexEnd)
					{
						iYear = _tstoi (strDate.Mid (m_iYearIndexStart, 4));
						bValid = ValidateLeapYear (iDay, iMonth, iYear);

						if (bValid)
						{
							bValid = iYear > 1752;

							if (!bValid && pstrErrorMsg)
								*pstrErrorMsg = _T("Years < 1753 not supported.");
						}
						else if (pstrErrorMsg)
							*pstrErrorMsg = _T("Invalid Leap Year");
					}
				}
				else
				{
					if (pstrErrorMsg)
						*pstrErrorMsg = _T("Invalid Day/Month Combination");
				}
			}
			else
			{
				if (pstrErrorMsg)
					*pstrErrorMsg = _T("Day must be 1-31");
			}
		}
	}

	return bValid;
}

bool CDateTimeEdit::IsYearMonthDayValid (const CString &strDate,
	CString *pstrErrorMsg/* = 0*/)
{
	bool bValid = true;
	const int iLen = strDate.GetLength ();
	int iDay = 0;
	int iMonth = 0;
	int iYear = 0;

	if (iLen >= 4)
	{
		iYear = _tstoi (strDate.Left (4));

		if (iLen > m_iMonthIndexStart)
		{
			if (iLen >= m_iMonthIndexEnd)
			{
				iMonth = _tstoi (strDate.Mid (m_iMonthIndexStart, 2));
			}
			else
			{
				iMonth = _tstoi (strDate.Mid (m_iMonthIndexStart));
			}

			// Month only test
			bValid = iLen == m_iMonthIndexEnd - 1 || (iMonth >= 1 && iMonth <= 12);

			if (!bValid)
			{
				if (pstrErrorMsg)
					*pstrErrorMsg = _T("Month must be 1-12");
			}
			else if (iLen > m_iDayIndexStart)
			{
				if (iLen < m_iDayIndexEnd)
				{
					iDay = _tstoi (strDate.Right (1));
				}
				else
				{
					iDay = _tstoi (strDate.Mid (m_iDayIndexStart, 2));
				}

				// Day only test
				bValid = iLen == m_iDayIndexEnd - 1 || (iDay >= 1 && iDay <= 31);

				if (bValid)
				{
					// Day and Month test
					bValid = ValidateDayMonth (iDay, iMonth);

					if (bValid)
					{
						bValid = ValidateLeapYear (iDay, iMonth, iYear);

						if (bValid)
						{
							bValid = iYear > 1752;

							if (!bValid && pstrErrorMsg)
								*pstrErrorMsg = _T("Years < 1753 not supported.");
						}
						else if (pstrErrorMsg)
							*pstrErrorMsg = _T("Invalid Leap Year");
					}
					else
					{
						if (pstrErrorMsg)
							*pstrErrorMsg = _T("Invalid Day/Month Combination");
					}
				}
				else
				{
					if (pstrErrorMsg)
						*pstrErrorMsg = _T("Day must be 1-31");
				}
			}
		}
	}

	return bValid;
}

// TODO: Review function.
void CDateTimeEdit::MoveLeft ()
{
	int iSelStart = 0;
	int iSelEnd = 0;

	// Note GetSel over-ridden
	GetSel (iSelStart, iSelEnd);
	FormatDateTime ();

	int iStart = 0;
	int iEnd = 0;
	// Calc which part we're in now.
	e_Part ePart = LocatePart (iSelStart, iStart, iEnd);

	if (iSelStart != iSelEnd || iEnd > GetWindowTextLength ())
	{
		const int iOffset = 1;
		const int iSpace = m_eMode == eDateTime && ePart == eHour ? 1 : 0;

		iSelStart = iStart - iOffset - iSpace;
		SetSel (iSelStart, iSelStart);
		// Calc where we're going next.
		ePart = LocatePart (iSelStart, iStart, iEnd);
	}

	SetSel (iStart, iEnd);
}

// TODO: Review function
void CDateTimeEdit::MoveRight ()
{
	CString strText;
	bool bFormatted = false;
	bool bSelectEnd = false;

	GetWindowText (strText);

	if (m_eMode != eTime)
	{
		TCHAR ch = strText.GetLength () ? strText[strText.GetLength () - 1] : 0;

		bFormatted = FormatDate (strText);
		bSelectEnd = strText.Right (1) != ch;

		if (ExpandYear (strText))
		{
			bFormatted |= FormatDate (strText);
		}
	}

	if (m_eMode != eDate)
	{
		TCHAR ch = strText.GetLength () ? strText[strText.GetLength () - 1] : 0;

		bFormatted |= FormatTime (strText);
		bSelectEnd |= strText.Right (1) != ch;
	}

	const int iLen = strText.GetLength ();
	int iSelStart = 0;
	int iSelEnd = 0;

	if (bFormatted && bSelectEnd)
	{
		iSelStart = iLen;
	}
	else
	{
		// Note GetSel over-ridden
		GetSel (iSelStart, iSelEnd);
	}

	int iStart = 0;
	int iEnd = 0;
	e_Part ePart = LocatePart (iSelStart, iStart, iEnd);

	if (m_eMode == eTime)
	{
		if (iLen == iStart)
		{
			COleDateTime dt = COleDateTime::GetCurrentTime ();
			CString strPart;

			switch (ePart)
			{
			case eHour:
				strPart.Format (_T("%02d"), dt.GetHour ());
				break;
			case eMinute:
				strPart.Format (_T("%02d"), dt.GetMinute ());
				break;
			case eSecond:
				strPart.Format (_T("%02d"), dt.GetSecond ());
				break;
			default:
				assert (0);
				break;
			}

			strText += strPart;
		}
		else
		{
			int iCurrStart = 0;
			int iCurrEnd = 0;

			// Note GetSel over-ridden
			GetSel (iCurrStart, iCurrEnd);

			if (iCurrStart != iCurrEnd)
			{
				const int iOffset = 1;
				const int iSelStart = iEnd + iOffset;

				SetSel (iSelStart, iSelStart);
				ePart = LocatePart (iSelStart, iStart, iEnd);
			}

			SetSel (iStart, iEnd);
		}
	}
	else
	{
		if (iLen == iStart)
		{
			COleDateTime dt = COleDateTime::GetCurrentTime ();
			CString strPart;

			switch (ePart)
			{
			case eDay:
				strPart.Format (_T("%02d"), dt.GetDay ());
				break;
			case eMonth:
				strPart.Format (_T("%02d"), dt.GetMonth ());
				break;
			case eYear:
				strPart.Format (_T("%04d"), dt.GetYear ());
				break;
			case eHour:
				strPart.Format (_T("%02d"), dt.GetHour ());
				break;
			case eMinute:
				strPart.Format (_T("%02d"), dt.GetMinute ());
				break;
			case eSecond:
				strPart.Format (_T("%02d"), dt.GetSecond ());
				break;
			default:
				assert (0);
				break;
			}

			strText += strPart;
		}
		else
		{
			if (iStart != iEnd)
			{
				const int iSelStart = iEnd + 1;

				SetSel (iSelStart, iSelStart);
				ePart = LocatePart (iSelStart, iStart, iEnd);
			}

			if (iLen >= iEnd)
			{
				SetSel (iStart, iEnd);
			}
			else
			{
				SetSel (iEnd, iEnd);
			}
		}
	}

	bool bProcess = true;

	if (m_eMode != eTime)
	{
		bProcess = IsDateValid (strText);

		if (!bProcess && ePart < eHour)
		{
			CString strPart;
			int i = 0;

			do
			{
				strPart = (strText.Mid (iStart, iEnd - iStart));
				i = _tstoi (strPart);

				switch (ePart)
				{
				case eDay:
				case eMonth:
					--i;
					strPart.Format(_T("%02d"), i);
					break;
				case eYear:
					--i;
					strPart.Format(_T("%04d"), i);
					break;
				default:
					assert (0);
					break;
				}

				strText.Delete (iStart, iEnd - iStart);
				strText.Insert (iStart, strPart);
				bProcess = IsDateValid (strText);
			} while (!bProcess);
		}
	}

	if (bProcess)
	{
		const bool bProcessChars = GetProcessChars ();

		SetProcessChars (false);
		SetWindowText (strText);
		SetProcessChars (bProcessChars);

		if (bSelectEnd)
			SetSel (strText.GetLength (), strText.GetLength ());
		else
			SetSel (iStart, iEnd);
	}
}

void CDateTimeEdit::SetTimeIndexes ()
{
	switch (m_eMode)
	{
	case eDate:
		m_iHourIndexStart = -1;
		m_iHourIndexEnd = -1;
		m_iMinuteIndexStart = -1;
		m_iMinuteIndexEnd = -1;
		m_iSecondIndexStart = -1;
		m_iSecondIndexEnd = -1;
		break;
	case eTime:
		m_iHourIndexStart = 0;
		m_iHourIndexEnd = 2;
		m_iMinuteIndexStart = m_iHourIndexEnd + 1;
		m_iMinuteIndexEnd = m_iMinuteIndexStart + 2;
		m_iSecondIndexStart = m_iMinuteIndexEnd + 1;
		m_iSecondIndexEnd = m_iSecondIndexStart + 2;
		break;
	case eDateTime:
		m_iHourIndexStart = m_iDateLength + 1;
		m_iHourIndexEnd = m_iHourIndexStart + 2;
		m_iMinuteIndexStart = m_iHourIndexEnd + 1;
		m_iMinuteIndexEnd = m_iMinuteIndexStart + 2;
		m_iSecondIndexStart = m_iMinuteIndexEnd + 1;
		m_iSecondIndexEnd = m_iSecondIndexStart + 2;
		break;
	default:
		assert (0);
		break;
	}
}

bool CDateTimeEdit::ValidateDayMonth (const int iDay, const int iMonth)
{
	const int iMaxDay = MaxDay (iMonth, -1);

	return iDay <= iMaxDay;
}

bool CDateTimeEdit::ValidateLeapYear (const int iDay, const int iMonth,
	const int iYear)
{
	bool bValid = true;

	// Day, Month and Year validation
	if (iDay == 29 && iMonth == 2)
	{
		if (iYear > 1752)
		{
			// Leap year calculation taken from K&R "The C Programming Language"
			bValid = iYear % 4 == 0 && iYear % 100 != 0 || iYear % 400 == 0;
		}
		else
		{
			bValid = false;
		}
	}

	return bValid;
}

bool CDateTimeEdit::ExpandYear (CString &strDate)
{
	bool bExpanded = false;
	CString strText (strDate);

	// Take care of case where format is DDMMYYYY or MMDDYYYY
	// and only two digits have been entered when tabbing out
	// of the control.
	if (m_chFieldOrder != 'y' && strText.GetLength () == m_iDateLength - 2)
	{
		const DWORD dwYear = _tstoi (strText.Mid (m_iYearIndexStart, 2));
		DWORD dwCenturyPrefix = m_dwCenturyPrefix;
		CString strPrefix;

		if (dwYear > m_dwTwoDigitYearMax)
		{
			dwCenturyPrefix--;
		}

		strPrefix.Format (_T("%02d"), dwCenturyPrefix);
		strText.Insert (m_iYearIndexStart, strPrefix);

		// In case a leap year check is needed...
		if (m_chFieldOrder == 'd' && IsDayMonthYearValid (strText) ||
			m_chFieldOrder == 'm' && IsMonthDayYearValid (strText))
		{
			strDate = strText;
			bExpanded = true;
		}
	}

	return bExpanded;
}

CDateTimeEdit::e_Part CDateTimeEdit::LocatePart (const int iSelStart, int &iStart, int &iEnd) const
{
	e_Part ePart = eDay;

	switch (m_eMode)
	{
	case eDate:
	{
		ePart = LocateDate (iSelStart, iStart, iEnd);
		break;
	}
	case eTime:
	{
		ePart = LocateTime (iSelStart, iStart, iEnd);
		break;
	}
	case eDateTime:
	{
		if (iSelStart <= m_iDateLength)
		{
			ePart = LocateDate (iSelStart, iStart, iEnd);
		}
		else
		{
			ePart = LocateTime (iSelStart, iStart, iEnd);
		}

		break;
	}
	default:
		assert (0);
		break;
	}

	return ePart;
}

CDateTimeEdit::e_Part CDateTimeEdit::LocateDate (const int iSelStart,
	int &iStart, int &iEnd) const
{
	e_Part ePart = eDay;

	if (iSelStart >= m_iDayIndexStart && iSelStart <= m_iDayIndexEnd)
	{
		iStart = m_iDayIndexStart;
		iEnd = m_iDayIndexEnd;
	}
	else if (iSelStart >= m_iMonthIndexStart && iSelStart <= m_iMonthIndexEnd)
	{
		ePart = eMonth;
		iStart = m_iMonthIndexStart;
		iEnd = m_iMonthIndexEnd;
	}
	else if (iSelStart >= m_iYearIndexStart && iSelStart <= m_iYearIndexEnd)
	{
		ePart = eYear;
		iStart = m_iYearIndexStart;
		iEnd = m_iYearIndexEnd;
	}

	return ePart;
}

CDateTimeEdit::e_Part CDateTimeEdit::LocateTime (const int iSelStart, int &iStart, int &iEnd) const
{
	e_Part ePart = eHour;

	if (iSelStart >= m_iHourIndexStart && iSelStart <= m_iHourIndexEnd)
	{
		iStart = m_iHourIndexStart;
		iEnd = m_iHourIndexEnd;
	}
	else if (iSelStart >= m_iMinuteIndexStart && iSelStart <= m_iMinuteIndexEnd)
	{
		ePart = eMinute;
		iStart = m_iMinuteIndexStart;
		iEnd = m_iMinuteIndexEnd;
	}
	else if (iSelStart >= m_iSecondIndexStart && iSelStart <= m_iSecondIndexEnd)
	{
		ePart = eSecond;
		iStart = m_iSecondIndexStart;
		iEnd = m_iSecondIndexEnd;
	}

	return ePart;
}

int CDateTimeEdit::GetDay (const CString &strDate)
{
	int iDay = -1;

	if (strDate.GetLength () >= m_iDayIndexEnd)
	{
		iDay = _tstoi (strDate.Mid (m_iDayIndexStart, 2));
	}

	return iDay;
}

int CDateTimeEdit::GetMonth (const CString &strDate)
{
	int iMonth = -1;

	if (strDate.GetLength () >= m_iMonthIndexEnd)
	{
		iMonth = _tstoi (strDate.Mid (m_iMonthIndexStart, 2));
	}

	return iMonth;
}

int CDateTimeEdit::GetYear (const CString &strDate)
{
	int iYear = -1;

	if (strDate.GetLength () >= m_iYearIndexEnd)
	{
		iYear = _tstoi (strDate.Mid (m_iYearIndexStart, 4));
	}

	return iYear;
}

int CDateTimeEdit::MaxDay (const int iMonth, const int iYear)
{
	int iDay = 31;

	if (iMonth > -1)
	{
		if (iMonth == 2)
		{
			iDay = 29;
		}
		else if (iMonth == 4 || iMonth == 6 || iMonth == 9 || iMonth == 11)
		{
			iDay = 30;
		}
		
		if (iYear > -1 && iMonth == 2)
		{
			if (!ValidateLeapYear (29, 2, iYear))
			{
				iDay = 28;
			}
		}
	}

	return iDay;
}

BEGIN_MESSAGE_MAP(CDateTimeEdit, CBaseEdit)
	//{{AFX_MSG_MAP(CDateTimeEdit)
	ON_WM_KEYDOWN()
	ON_WM_CHAR()
	ON_WM_LBUTTONDOWN()
	ON_WM_LBUTTONUP()
	ON_WM_LBUTTONDBLCLK()
	ON_WM_MOUSEMOVE()
	ON_WM_KILLFOCUS()
	ON_BN_CLICKED(123, OnCalendar)
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

void CDateTimeEdit::PreSubclassWindow ()
{
	CRect ClientRect;
	CRect ButtonRect;
	DWORD dwStyle = BS_BITMAP | WS_CHILD | WS_TABSTOP;

	m_dwMargins = GetMargins ();
	GetClientRect (ClientRect);

	if (m_eMode != eTime)
	{
		SetMargins (LOWORD(m_dwMargins), ClientRect.bottom - ClientRect.top);
		dwStyle |= WS_VISIBLE;
	}

	if (!IsWindowEnabled () || (GetStyle () & ES_READONLY))
	{
		dwStyle |= WS_DISABLED;
	}

	ButtonRect.SetRect (ClientRect.right - (ClientRect.bottom - ClientRect.top),
		0, ClientRect.right, ClientRect.bottom);
	m_ButtonCalendar.Create (_T(""), dwStyle, ButtonRect, this, 123);
	m_ButtonCalendar.SetBitmap (static_cast<HBITMAP>(::LoadImage
		(AfxGetResourceHandle (), MAKEINTRESOURCE(IDB_CALENDAR), IMAGE_BITMAP,
		13, 12, LR_DEFAULTCOLOR)));
}

LRESULT CDateTimeEdit::WindowProc (UINT message, WPARAM wParam, LPARAM lParam)
{
	const LRESULT lRes = m_bFromSpin && message == EM_SETSEL ? 0 :
		CBaseEdit::WindowProc (message, wParam, lParam);

	switch (message)
	{
	case WM_ENABLE:
		m_ButtonCalendar.EnableWindow (wParam);
		break;
	case EM_SETSEL:
		m_bFromSpin = false;
		m_ButtonCalendar.Invalidate ();
		break;
	case EM_SETREADONLY:
		m_ButtonCalendar.EnableWindow (!wParam);
		break;
	default:
		break;
	}

	return lRes;
}

void CDateTimeEdit::OnKeyDown (UINT nChar, UINT nRepCnt, UINT nFlags)
{
	if (!(::GetKeyState (VK_CONTROL) & 0x80) &&
		!(::GetKeyState (VK_SHIFT) & 0x80) && (nChar == VK_LEFT ||
		nChar == VK_RIGHT))
	{
		ResetValid ();

		switch (nChar)
		{
		case VK_LEFT:
			MoveLeft ();
			break;
		case VK_RIGHT:
			MoveRight ();
			break;
		default:
			assert (0);
			break;
		}
	}
	else if ((::GetKeyState (VK_CONTROL) & 0x80) && nChar == VK_DOWN)
	{
		OnCalendar ();
	}
	else
	{
		CBaseEdit::OnKeyDown (nChar, nRepCnt, nFlags);
	}

	m_ButtonCalendar.Invalidate ();
}

// TODO: Review function.
void CDateTimeEdit::OnChar (UINT nChar, UINT nRepCnt, UINT nFlags)
{
	bool bFinished = false;
	const int iCurrLen = GetWindowTextLength ();
	int iSelStart = -1;
	int iSelEnd = -1;

	// Note GetSel over-ridden
	GetSel (iSelStart, iSelEnd);

	if (nChar != VK_BACK && (iSelStart != iSelEnd &&
		iSelEnd != iCurrLen))
	{
		int iStart = -1;
		int iEnd = -1;
		bool bOK = true;

		e_Part ePart = LocatePart (iSelStart, iStart, iEnd);

		if (iSelStart == iStart && iSelEnd == iEnd)
		{
			switch (ePart)
			{
			case eDay:
				bOK = nChar >= '0' && nChar <= '3';

				if (!bOK)
					DisplayToolTip (true, _T("Day must start with 0-3"));
				break;
			case eMonth:
				bOK = nChar >= '0' && nChar <= '1';

				if (!bOK)
					DisplayToolTip (true, _T("Month must start with 0 or 1"));
				break;
			case eYear:
				bOK = nChar >= '0' && nChar <= '9';
				break;
			case eHour:
				bOK = nChar >= '0' && nChar <= '2';

				if (!bOK)
					DisplayToolTip (true, _T("Hour must start with 0-2"));
				break;
			case eMinute:
				bOK = nChar >= '0' && nChar <= '5';

				if (!bOK)
					DisplayToolTip (true, _T("Minute must start with 0-5"));
				break;
			case eSecond:
				bOK = nChar >= '0' && nChar <= '5';

				if (!bOK)
					DisplayToolTip (true, _T("Second must start with 0-5"));
				break;
			}

			if (bOK)
			{
				const bool bProcessChars = GetProcessChars ();

				SetProcessChars (false);
				CBaseEdit::OnChar (nChar, nRepCnt, nFlags);
				SetProcessChars (bProcessChars);
			}

			bFinished = true;
		}
	}
	else if (iSelStart >= m_iYearIndexStart && iSelStart <= m_iYearIndexEnd &&
		iSelStart < iCurrLen)
	{
		if (nChar >= '0' && nChar <= '9')
		{
			CString strText;
			const TCHAR *pszText = 0;
			int iCount = 0;

			ValidateChar (nChar, false);
			strText = GetProposedText ();
			pszText = static_cast<const TCHAR *>(strText) + m_iYearIndexStart;

			while (*pszText >= '0' && *pszText <= '9')
			{
				++iCount;
				++pszText;
			}

			if (iCount <= 4)
			{
				const bool bProcessChars = GetProcessChars ();

				if (iCount == 4)
				{
					if (_tstoi (strText.Mid (m_iYearIndexStart, 4)) < 1752)
					{
						strText.Delete (m_iYearIndexStart, 4);
						strText.Insert (m_iYearIndexStart, _T("1753"));
					}
				}

				SetProcessChars (false);
				SetWindowText (strText);
				SetProcessChars (bProcessChars);
				SetSel (iSelStart + 1, iSelStart + 1);
				bFinished = true;
			}
		}
	}
	else if (nChar == ' ' && iCurrLen == 0)
	{
		const COleDateTime dt (COleDateTime::GetCurrentTime ());
		CString strToday;

		switch(m_eMode)
		{
		case eDate:
			strToday = dt.Format (VAR_DATEVALUEONLY);
			break;
		case eTime:
			strToday = dt.Format (VAR_TIMEVALUEONLY);
			break;
		case eDateTime:
			strToday = dt.Format ();
			break;
		default:
			assert (0);
			break;
		}

		const int iLen = strToday.GetLength ();
		const bool bProcessChars = GetProcessChars ();

		SetProcessChars (false);
		SetWindowText (strToday);
		SetProcessChars (bProcessChars);
		SetSel (iLen, iLen);
		bFinished = true;
	}

	if (!bFinished)
	{
		CString strText;
		bool bFormatted = false;

		// Use this function to generate the 'proposed text'.
		// Don't check return value, because we will re-format first.
		ValidateChar (nChar);
		strText = GetProposedText ();

		if (nChar != VK_BACK)
		{
			if (m_eMode != eTime)
			{
				/// Pre-process
				bFormatted = FormatDate (strText);
			}
		}

		// Syntax check
		if (Match (strText))
		{
			CString strErrorMsg;

			if (SemanticCheck (strText, &strErrorMsg))
			{
				if (nChar != VK_BACK)
				{
					if (m_eMode != eDate)
					{
						bFormatted |= FormatTime (strText);
					}
				}

				if (bFormatted)
				{
					const int iLen = strText.GetLength ();
					const bool bProcessChars = GetProcessChars ();

					SetProcessChars (false);
					SetWindowText (strText);
					SetProcessChars (bProcessChars);
					SetSel (iLen, iLen);
				}
				else
				{
					CBaseEdit::OnChar (nChar, nRepCnt, nFlags);
				}
			}
			else
			{
				DisplayToolTip (true, strErrorMsg);
			}
		}
	}

	m_ButtonCalendar.Invalidate ();
}

void CDateTimeEdit::OnLButtonDown (UINT nFlags, CPoint point)
{
	FormatDateTime ();
	CBaseEdit::OnLButtonDown (nFlags, point);
	m_ButtonCalendar.Invalidate ();
}

// TODO: Review function.
void CDateTimeEdit::OnLButtonUp (UINT nFlags, CPoint point)
{
	int iSelStart = 0;
	int iSelEnd = 0;

	CBaseEdit::OnLButtonUp (nFlags, point);
	// Note GetSel over-ridden
	GetSel (iSelStart, iSelEnd);

	int iStart = 0;
	int iEnd = 0;
	const e_Part ePart = LocatePart (iSelStart, iStart, iEnd);
	const int iLen = GetWindowTextLength ();

	if (iSelStart <= iLen && iSelEnd < iLen)
	{
		SetSel (iStart, iEnd);
	}

	m_ButtonCalendar.Invalidate ();
}

void CDateTimeEdit::OnLButtonDblClk (UINT nFlags, CPoint point)
{
	CBaseEdit::OnLButtonDblClk (nFlags, point);
	m_ButtonCalendar.Invalidate ();
}

void CDateTimeEdit::OnMouseMove (UINT nFlags, CPoint point)
{
	CBaseEdit::OnMouseMove (nFlags, point);
	m_ButtonCalendar.Invalidate ();
}

void CDateTimeEdit::OnKillFocus (CWnd *pNewWnd)
{
	CString strText;

	GetWindowText (strText);

	const int iLen = strText.GetLength ();

	if (m_eMode != eTime)
	{
		ExpandYear (strText);

		if (m_eMode == eDateTime)
		{
			if (iLen == m_iDateLength)
			{
				strText += " 00:00:00";
			}
			else if (iLen == m_iDateLength + 1)
			{
				strText += "00:00:00";
			}
		}
	}

	if (m_eMode != eDate)
	{
		if (iLen == m_iSecondIndexStart)
		{
			strText += "00";
		}
	}

	SetSel (0, -1);
	ReplaceSel (strText, TRUE);
	CBaseEdit::OnKillFocus (pNewWnd);
}

void CDateTimeEdit::OnCalendar ()
{
	const int iDay = GetDay ();
	const int iMonth = GetMonth ();
	const int iYear = GetYear ();

	// m_pCalendarDlg deleted/zeroed by CCalendarDlg
	m_pCalendarDlg = new CCalendarDlg (this, this);
	m_pCalendarDlg->Create (IDD_CALENDAR, this);
	m_pCalendarDlg->SetDate (iDay, iMonth, iYear);
	m_pCalendarDlg->ShowWindow (TRUE);
}
}
